package AccountManager::WebRequest;

use strict;
use warnings;

use CGI;
use DateTime;
use English qw(-no_match_vars);
use Template;
use Log::Any::Adapter;
use List::MoreUtils qw(uniq);
use HTTP::AcceptLanguage;
use HTTP::Message;
use HTTP::Response;
use IO::String;
use Text::CSV;

use AccountManager::Account;
use AccountManager::Account::Manager;
use AccountManager::Metadata;
use AccountManager::Service;
use AccountManager::Token;
use AccountManager::Tools;

## Defining parameters format
my $urn_or_url_regex = '(http(s?):\/\/|urn:)[^\\\$\*\"\'\`\^\|\<\>\n\s]+'
  ;    ## Format de type URL HTTP ou URN
my $url_regex     = 'http(s?):\/\/[^\\\$\*\"\'\`\^\|\<\>\n\s]+';
my $email_regex   = '([\w\-\_\.\/\+\=\'\&]+|\".*\")\@[\w\-]+(\.[\w\-]+)+';
my $domains_regex = '[\w\.\-]+(,[\w\.\-]+)*';
my %format        = (
    ## URL
    #'attributeauthority' => $url_regex,
    'entityid' => $urn_or_url_regex,
);

my %actions = (
    home               => 'req_home',
    select_sp          => 'req_select_sp',
    select_email       => 'req_select_email',
    complete_challenge => 'req_complete_challenge',
    create_accounts    => 'req_create_accounts',
);

## New web request
sub new {
    my ($pkg, %args) = @_;

    my $self = {
        configuration => $args{configuration},
    };

    Log::Any::Adapter->set(
        'File',
        $self->{configuration}->{log}->{file},
        log_level => $self->{configuration}->{log}->{level}
    );

    $self->{logger} = Log::Any->get_logger();

    AccountManager::DB->register_db(
        driver          => $self->{configuration}->{database}->{type},
        database        => $self->{configuration}->{database}->{name},
        host            => $self->{configuration}->{database}->{host},
        password        => $self->{configuration}->{database}->{password},
        username        => $self->{configuration}->{database}->{user}
    );

    $self->{db} = AccountManager::DB->new();
    $self->{cgi} = CGI->new();

    bless $self, $pkg;

    return $self;
}

sub run {
    my ($self) = @_;


    # process input parameters
    my %parameters = $self->{cgi}->Vars();

    foreach my $parameter (keys %parameters) {

        # cleanup
        $parameters{$parameter} =~ s/\r//g;  # remove &0D char
        $parameters{$parameter} =~ s/\s+$//; # remove trailing spaces
        $parameters{$parameter} =~ s/^\s+//; # remove leading spaces

        # format check
        if (defined $format{$parameter}
            && !ref($format{$parameter})) {
            if ($parameters{$parameter} !~ /^$format{$parameter}$/) {
                $self->{logger}->error(
                    "Incorrect parameter format : $parameter"
                );
                $self->respond({ errors => [ "format_$parameter" ] });
            }
        }

        # If action_xx parameter is set, set action parameter with value xx
        if ($parameter =~ /^action_(\w+)$/) {
            $parameters{action} = $1;
        }

        # register needed parameters
        $self->{in} = {
            email    => $parameters{email},
            style    => $parameters{style},
            entityid => $parameters{entityid},
            token    => $parameters{token}
        };
    }

    # process requested action
    my $action = $parameters{action} || 'home';
    if ($actions{$action}) {
        my $method = $actions{$action};
        $self->$method();
    } else {
        ## unknown action
        $self->{logger}->error( "Unknown action '$action'");
        $self->respond({ errors => [ "Unknown action '$action'" ] });
    }

    return 1;
}

## Return HTML content
sub respond {
    my ($self, $data) = @_;
    
    $data->{app} = {
        name          => $self->{configuration}->{app}->{name},
        url           => $self->{configuration}->{app}->{url},
        support_email => $self->{configuration}->{app}->{support_email},
        version       => $self->{configuration}->{app}->{version},
    };

    my $lang = HTTP::AcceptLanguage->new($ENV{HTTP_ACCEPT_LANGUAGE})->match(qw/en fr/) || 'en';

    ## Parse template
    my $tt2 = Template->new({
        INCLUDE_PATH => $self->{configuration}->{_}->{templates_dir} . "/web/$lang"
    });

    my $template;

    ## nobanner is used to do AJAX to get only pieces of HTML to load in the web client
    if ($self->{in}->{style} && $self->{in}->{style} eq 'nobanner') {
        $template = 'index-nobanner.tt2.html';
    } else {
        $template = 'index.tt2.html';
    }

    binmode(STDOUT, ":utf8");

    print $self->{cgi}->header(
        -nph     => 1,
        -type    => 'text/html',
        -charset => 'utf8'
    );

    unless ($tt2->process($template, $data, \*STDOUT)) {
        printf "Content-type: text/plain\n\n Error: %s", $tt2->error();
        $self->{logger}->errorf("Web parser error : %s", $tt2->error());
    }

}

sub req_select_sp {
    my ($self) = @_;

    my $metadata;

    eval {
        $metadata = AccountManager::Metadata->new(
            file => $self->{configuration}->{_}->{federation_metadata_file}
        );
    };
    if ($EVAL_ERROR) {
        $self->{logger}->error("Failed to load federation metadata: $EVAL_ERROR");
        $self->respond({ errors => [ "internal" ] });
    }

    $self->respond({
        env => {
            SCRIPT_NAME => $ENV{SCRIPT_NAME}
        },
        metadata => $metadata->parse(type => 'sp'),
        action   => 'select_sp'
    });
}

sub req_select_email {
    my ($self) = @_;

    unless ($self->{in}->{entityid}) {
        $self->{logger}->error("Missing parameter entityid");
        $self->respond({ errors => [ "missing_entityid" ] });
    }

    # Create a persistent service provider object
    my $provider = AccountManager::Service->new(
        db       => $self->{db},
        entityid => $self->{in}->{entityid}
    );

    if ($provider->load(speculative => 1)) {
        # already present in DB, nothing todo
    } else {
        # extract information from metadata
        my $metadata;

        eval {
            $metadata = AccountManager::Metadata->new(
                file => $self->{configuration}->{_}->{federation_metadata_file}
            );
        };
        if ($EVAL_ERROR) {
            $self->{logger}->error("Failed to load federation metadata: $EVAL_ERROR");
            $self->respond({ errors => [ "internal" ] });
        }

        my $sps = $metadata->parse(id => $self->{in}->{entityid});
        if (!@$sps) {
            $self->{logger}->errorf(
                "No such SP '%s' in metadata", $self->{in}->{entityid}
            );
            $self->respond({ errors => [ "no_such_entity" ] });
        }
        my $sp = $sps->[0];

        # complete persistent object
        $provider->displayname($sp->{display_name});
        $provider->contacts(uniq map { $_->{EmailAddress} } @{$sp->{contacts}})
            if $sp->{contacts};

        # save in DB
        unless ($provider->save()) {
            $self->{logger}->error("Failed to save service provider object");
            $self->respond({ errors => [ "internal" ] });
        }
    }

    # override metadata contacts if needed
    my $entity = $self->{in}->{entityid};
    my $contacts =
        $self->{configuration}->{$entity}->{contacts} ||
        $self->{configuration}->{service}->{contacts};
    if ($contacts) {
        if ($contacts =~ /^\+(.+)/) {
            # complement original contacts
            $provider->contacts($provider->contacts(), split(/, */, $1));
        } else {
            # replace original contacts
            $provider->contacts(split(/, */, $contacts));
        }
    }

    $self->respond({
        provider => $provider,
        action   => 'select_email'
    });
}

sub req_complete_challenge {
    my ($self) = @_;

    unless ($self->{in}->{entityid}) {
        $self->{logger}->error("Missing parameter entityid");
        $self->respond({ errors => [ "missing_entityid" ] });
    }

    unless ($self->{in}->{email}) {
        $self->{logger}->error("Missing parameter email");
        $self->respond({ errors => [ "missing_email" ] });
    }

    my $provider = AccountManager::Service->new(
        db       => $self->{db},
        entityid => $self->{in}->{entityid},
    );

    unless ($provider->load(speculative => 1)) {
        $self->{logger}->errorf("No such SP '%s' in database", $self->{in}->{entityid});
        $self->respond({ errors => [ "no_such_entity" ] });
    }

    # override metadata contacts if needed
    my $entity = $self->{in}->{entityid};
    my $contacts =
        $self->{configuration}->{$entity}->{contacts} ||
        $self->{configuration}->{service}->{contacts};
    if ($contacts) {
        if ($contacts =~ /^\+(.+)/) {
            # complement original contacts
            $provider->contacts($provider->contacts(), split(/, */, $1));
        } else {
            # replace original contacts
            $provider->contacts(split(/, */, $contacts));
        }
    }

    ## Check that email is a known contact for this SP
    unless ($provider->is_contact($self->{in}->{email}))
    {
        $self->{logger}->errorf(
            "Requested a token for %s for an unautorized address '%s'",
            $self->{in}->{entityid},
            $self->{in}->{email}
        );
        $self->respond({ errors => [ "internal" ] });
    }

    # delete any previous token for the same email/service couple
    my $old_token = AccountManager::Token->new(
        db            => $self->{db},
        email_address => $self->{in}->{email},
        sp_entityid   => $self->{in}->{entityid}
    );

    if ($old_token->load(speculative => 1)) {
        unless ($old_token->delete()) {
            $self->{logger}->errorf(
                "Failed to delete previous authentication token with ID %s",
                $old_token->id()
            );
            $self->respond({ errors => [ "internal" ] });
        }
    }

    # compute a new token
    my $validity_period =
        $self->{configuration}->{_}->{tokens_validity_period};
    my $token = AccountManager::Token->new(
        db              => $self->{db},
        email_address   => $self->{in}->{email},
        sp_entityid     => $self->{in}->{entityid},
        creation_date   => DateTime->now(),
        expiration_date => DateTime->now()->add(hours => $validity_period),
        token           => AccountManager::Tools::generate_secret(20)
    );

    unless ($token->save()) {
        $self->{logger}->error("Failed to save authentication token");
        $self->respond({ errors => [ "internal" ] });
    }

    my $sender    = $self->{configuration}->{_}->{notice_from};
    my $recipient = $self->{in}->{email};
    my $sendmail  = $self->{configuration}->{_}->{sendmail_path} ||
                    '/usr/sbin/sendmail';

    open(my $handle, '|-', "$sendmail -f $sender $recipient") or do {
        $self->{logger}->errorf("Unable to run sendmail executable: %s", $ERRNO);
        $self->respond({ errors => [ "mail_notification_error" ] });
    };

    my $tt2 = Template->new({
        INCLUDE_PATH => $self->{configuration}->{_}->{templates_dir}
    });
    my $template = 'mail/send_authentication_token.tt2.eml';
    my $data = {
        env => {
            REMOTE_ADDR => $ENV{REMOTE_ADDR},
        },
        conf => {
            app_name      => $self->{configuration}->{app}->{name},
            app_url       => $self->{configuration}->{app}->{url},
            support_email => $self->{configuration}->{app}->{support_email},
        },
        from     => $sender,
        to       => $recipient,
        entityid => $self->{in}->{entityid},
        token    => $token->token(),
    };

    unless ($tt2->process($template, $data, $handle)) {
        $self->{logger}->errorf("Mail notification error: %s", $tt2->error());
        $self->respond({ errors => [ "mail_notification_error" ] });
    }

    close $handle;

    $self->{logger}->infof(
        "Token send to %s for entityid=%s;token=%s",
        $self->{in}->{email},
        $self->{in}->{entityid},
        $token->token(),
    );

    $self->respond({
        email    => $self->{in}->{email},
        entityid => $self->{in}->{entityid},
        action   => 'complete_challenge'
    });
}

sub req_create_accounts {
    my ($self) = @_;

    unless ($self->{in}->{entityid}) {
        $self->{logger}->error("Missing parameter entityid");
        $self->respond({ errors => [ "missing_entityid" ] });
    }

    unless ($self->{in}->{token}) {
        $self->{logger}->error("Missing parameter token");
        $self->respond({ errors => [ "missing_token" ] });
    }

    my $token = AccountManager::Token->new(
        db    => $self->{db},
        token => $self->{in}->{token}
    );

    if (! $token->load(speculative => 1)) {
        $self->{logger}->errorf(
            "Failed to validate authentication token %s for entityid %s",
            $self->{in}->{token},
            $self->{in}->{entityid}
        );
        $self->respond({ errors => [ "wrong_token" ] });
    }

    if (! $token->sp_entityid() eq $self->{in}->{entityid}) {
        $self->{logger}->errorf(
            "Authentication token %s cannot be used for SP with entityid %s",
            $self->{in}->{token},
            $self->{in}->{entityid}
        );
        $self->respond({ errors => [ "wrong_token_for_sp" ] });
    }

    ## delete the token
    unless ($token->delete()) {
        $self->{logger}->errorf(
            "Failed to delete authentication token %s",
            $self->{in}->{token}
        );
    }

    ## create test accounts
    my @accounts;

    my $entity = $self->{in}->{entityid};
    my $profiles =
        $self->{configuration}->{$entity}->{account_profiles} ||
        $self->{configuration}->{service}->{account_profiles};
    my $validity_period =
        $self->{configuration}->{$entity}->{account_validity_period} ||
        $self->{configuration}->{service}->{account_validity_period};

    foreach my $profile (split(/, */, $profiles)) {
        my $password = AccountManager::Tools::generate_password(10);
        my $account = AccountManager::Account->new(
            db              => $self->{db},
            profile         => $profile,
            sp_entityid     => $entity,
            scope           => $self->{configuration}->{idp}->{scope},
            password        => $password,
            password_hash   => AccountManager::Tools::sha256_hash($password),
            creation_date   => DateTime->now(),
            expiration_date => DateTime->now()->add(days => $validity_period)
        );
        next unless $account->save();
        push @accounts, $account;
    }

    unless (@accounts) {
        $self->{logger}->errorf(
            "Failed to create test accounts for SP with entityid %s",
            $self->{in}->{entityid}
        );
        $self->respond({ errors => [ "accounts_creation_failed" ] });
    }

    ## Update simpleSAMLphp configuration to enable test accounts
    my $accounts = AccountManager::Account::Manager->get_accounts(
        db => $self->{db}
    );

    eval {
        AccountManager::Tools::update_ssp_authsources(
            $self->{configuration}->{_}->{templates_dir},
            $self->{configuration}->{idp}->{accounts_file},
            $accounts
        );
    };
    if ($EVAL_ERROR) {
        $self->{logger}->errorf(
            "Failed to create simpleSAMLphp configuration file: %s",
            $EVAL_ERROR
        );
        $self->respond({ errors => [ "accounts_creation_failed" ] });
    }

    $self->{logger}->infof(
        "Token validated for entityid=%s;token=%s",
        $self->{in}->{entityid},
        $self->{in}->{token}
    );

    binmode(STDOUT, ":utf8");

    my $response = HTTP::Response->new(
        200, 'OK', [ 'Content-Type' => 'multipart/x-mixed-replace' ]
    );

    $response->protocol('HTTP/1.1');
    $response->date(time);
    $response->server($ENV{SERVER_SOFTWARE});

    # HTML page
    my $data = {
        app => {
            name          => $self->{configuration}->{app}->{name},
            url           => $self->{configuration}->{app}->{url},
            support_email => $self->{configuration}->{app}->{support_email},
            version       => $self->{configuration}->{app}->{version},
        },
        accounts                 => \@accounts,
        accounts_validity_period => $self->{configuration}->{service}->{account_validity_period},
        idp_displayname          => $self->{configuration}->{idp}->{displayname},
        entityid                 => $self->{in}->{entityid},
        action                   => 'create_accounts'
    };

    my $lang = HTTP::AcceptLanguage->new($ENV{HTTP_ACCEPT_LANGUAGE})->match(qw/en fr/) || 'en';

    my $tt2 = Template->new({
        INCLUDE_PATH => $self->{configuration}->{_}->{templates_dir} . "/web/$lang"
    });

    my $page_content;
    $tt2->process('index.tt2.html', $data, \$page_content);

    my $page_response = HTTP::Message->new(
        [
            'Content-Type'        => 'text/html; charset=utf-8',
            'Content-Disposition' => 'inline'
        ],
        $page_content
    );
    $response->add_part($page_response);

    # CSV file
    my $csv = Text::CSV->new({ binary => 1, eol => "\r\n", quote_space => 0 });
    my $file_content;
    my $file_content_io = IO::String->new($file_content);
    $csv->print($file_content_io, [ qw/
        username
        password
        profile
        cn
        displayName
        givenName
        mail
        eduPersonAffiliation
        eduPersonScopedAffiliation
        eduPersonPrincipalName
        schacHomeOrganization
        schacHomeOrganizationType
    / ]);

    foreach my $account (@accounts) {
        $csv->print($file_content_io, [
            $account->internal_uid(),
            $account->password(),
            $account->profile(),
            $account->cn(),
            $account->displayName(),
            $account->givenName(),
            $account->mail(),
            $account->eduPersonAffiliation(),
            $account->eduPersonScopedAffiliation(),
            $account->eduPersonPrincipalName(),
            $account->schacHomeOrganization(),
            $account->schacHomeOrganizationType(),
        ]);
    }

    my $file_response = HTTP::Message->new(
        [
            'Content-Type'        => 'text/csv; charset=utf-8',
            'Content-Disposition' => 'attachment; filename="accounts.csv"'
        ],
        $file_content
    );
    $response->add_part($file_response);

    print $response->as_string();
}

## Return the homepage of the service
sub req_home {
    my ($self) = @_;

    $self->respond({
        action => 'home'
    });
}

1;
